from sklearn.cluster import KMeans
from sklearn.decomposition import FastICA, PCA
from sklearn.feature_extraction.text import CountVectorizer, TfidfVectorizer
from sklearn.impute import SimpleImputer
from sklearn.linear_model import LinearRegression, LogisticRegression, Lasso, Ridge, Perceptron
from sklearn.mixture import GaussianMixture
from sklearn.naive_bayes import GaussianNB
from sklearn.neighbors import KNeighborsRegressor, KNeighborsClassifier
from sklearn.neural_network import MLPClassifier, MLPRegressor
from sklearn.preprocessing import StandardScaler, LabelEncoder, OneHotEncoder
from sklearn.svm import LinearSVC
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor


def init_partializer(sklearn_piskle_partializer):
    sklearn_piskle_partializer.register_class_attributes(LinearRegression, ['coef_', 'intercept_'])
    sklearn_piskle_partializer.register_class_attributes(LogisticRegression, ['classes_', 'coef_', 'intercept_'])
    sklearn_piskle_partializer.register_class_attributes(DecisionTreeClassifier,
                                                         ['n_features_', 'n_outputs_', 'classes_',
                                                          'tree_'])
    sklearn_piskle_partializer.register_class_attributes(Perceptron, ['coef_', 'intercept_', 'classes_'])
    sklearn_piskle_partializer.register_class_attributes(Ridge, ['coef_', 'intercept_'])
    sklearn_piskle_partializer.register_class_attributes(Lasso, ['coef_', 'intercept_'])
    sklearn_piskle_partializer.register_class_attributes(GaussianMixture,
                                                         ['weights_', 'means_', 'precisions_cholesky_'])
    sklearn_piskle_partializer.register_class_attributes(GaussianNB, ['classes_', 'theta_', 'sigma_', 'class_prior_'])
    sklearn_piskle_partializer.register_class_attributes(KNeighborsClassifier,
                                                         ['outputs_2d_', 'classes_', '_y', 'effective_metric_',
                                                          '_fit_method', 'n_samples_fit_', '_tree'])
    sklearn_piskle_partializer.register_class_attributes(KNeighborsRegressor,
                                                         ['_y', 'effective_metric_', '_fit_method', 'n_samples_fit_',
                                                          '_tree'])
    sklearn_piskle_partializer.register_class_attributes(LinearSVC, ['classes_', 'coef_', 'intercept_'])
    sklearn_piskle_partializer.register_class_attributes(DecisionTreeRegressor, ['n_outputs_', 'tree_'])
    sklearn_piskle_partializer.register_class_attributes(MLPClassifier, ['_label_binarizer', 'n_outputs_', 'n_layers_',
                                                                         'out_activation_', 'coefs_', 'intercepts_'])
    sklearn_piskle_partializer.register_class_attributes(MLPRegressor,
                                                         ['n_layers_', 'out_activation_', 'coefs_', 'intercepts_'])
    sklearn_piskle_partializer.register_class_attributes(SimpleImputer, ['statistics_'])
    sklearn_piskle_partializer.register_class_attributes(StandardScaler, ['mean_', 'scale_'])
    sklearn_piskle_partializer.register_class_attributes(FastICA, ['components_', 'mean_'])
    sklearn_piskle_partializer.register_class_attributes(PCA, ['mean_', 'components_'])
    sklearn_piskle_partializer.register_class_attributes(KMeans, ['_n_threads', 'cluster_centers_'])
    sklearn_piskle_partializer.register_class_attributes(CountVectorizer, ['vocabulary_'])
    sklearn_piskle_partializer.register_class_attributes(TfidfVectorizer, ['_tfidf', 'vocabulary_'])

    sklearn_piskle_partializer.register_class_attributes(LabelEncoder, ['classes_'])
    sklearn_piskle_partializer.register_class_attributes(OneHotEncoder, ['categories_', 'drop_idx_'])
    return sklearn_piskle_partializer


# List of all modules:
# https://scikit-learn.org/stable/modules/classes.html


from piskle.piskle import Pisklizer
from piskle.sklearn_piskle_partializer import SklearnPisklePartializer
from piskle._piskle_initializer import init_partializer as _init_partializer

sklearn_piskle_partializer = _init_partializer(SklearnPisklePartializer())
sklearn_exporter = Pisklizer(sklearn_piskle_partializer)

__version__ = '0.0.1'


def dump(model, file, *args, **kwargs):
    return sklearn_exporter.dump(model, file, *args, **kwargs)


def load(file):
    return sklearn_exporter.load(file)



from sklearn import __version__ as sklearn_version

from piskle.piskle import PisklePartializer


class SklearnPisklePartializer(PisklePartializer):
    def get_default_params_dict(self, obj, version, params_dict):
        return obj.get_params() if params_dict is None else params_dict

    def register_class_attributes(self, class_, attributes, version=None):
        version = sklearn_version if version is None else version
        return super().register_class_attributes(class_, attributes, version)

    def get_default_version(self, obj, version):
        return sklearn_version



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="piskle",
    version="0.0.1",
    author="Amal HASNI, Dhia HMILA",
    author_email="emerald.snippets@gmail.com",
    description="Piskle allows you to selectively and efficiently serialize scikit-learn models "
                "to save on memory and load times.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/hmiladhia/piskle",
    packages=['piskle'],
    install_requires=['scikit-learn==0.24.0'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



from collections import defaultdict
from functools import partial

from piskle.utils import get_object_class_name, import_object


class PartialObject(dict):
    pass


class PisklePartializer:
    def __init__(self):
        self.class_attrs_dict = defaultdict(dict)
        self.custom_partializer_dict = defaultdict(dict)

    def from_partial_obj(self, partial_obj: PartialObject):
        # TODO: warning on sklearn version metadata mismatch
        # Getting the estimator class
        obj_class = import_object(*partial_obj['class_name'])

        # Instantiating the estimator
        model = obj_class(**partial_obj['params'])

        # Assigning the attributes to the estimator
        for attribute_name, attribute_value in partial_obj['attributes'].items():
            setattr(model, attribute_name, attribute_value)

        return model

    def to_partial_obj(self, obj, params_dict=None, version=None) -> PartialObject:
        version = self.get_default_version(obj, version) if version is None else version

        # Selecting partializer and obj_attributes
        partialize = self.get_partializer(obj, version)

        if partialize is None:
            return obj

        params_dict = self.get_default_params_dict(obj, version, params_dict)

        return partialize(obj, params_dict, version)

    def get_partializer(self, obj, version):
        obj_class = type(obj)
        if obj_class in self.custom_partializer_dict:
            return self.custom_partializer_dict[obj_class][version]
        elif obj_class in self.class_attrs_dict:
            return partial(self.default_partializer, obj_attributes=self.class_attrs_dict[obj_class][version])
        return None

    def get_default_params_dict(self, obj, version, params_dict):
        return params_dict

    def get_default_version(self, obj, version):
        return version

    def default_partializer(self, obj, params_dict=None, version=None, obj_attributes=None):
        obj_attributes = [] if obj_attributes is None else obj_attributes

        # TODO: include module version: in meta or in version ?

        # Getting the attributes
        model_vars = vars(obj)
        attributes_dict = {attr: model_vars[attr] for attr in obj_attributes}

        # Creating the partial object
        partial_obj = PartialObject(
            class_name=get_object_class_name(obj),
            attributes=attributes_dict,
            params=params_dict
        )
        return partial_obj

    def register_class_attributes(self, class_, attributes, version):
        self.class_attrs_dict[class_][version] = attributes
